package org.rabix.executor.execution;

import java.lang.Thread.UncaughtExceptionHandler;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.atomic.AtomicLong;

/**
 * Simple thread factory used to create {@link JobHandlerRunnable} threads.
 */
public class JobHandlerThreadFactoryBuilder {
  
  private String namePrefix = null;
  private boolean daemon = false;
  private int priority = Thread.NORM_PRIORITY;
  private ThreadFactory backingThreadFactory = null;
  private UncaughtExceptionHandler uncaughtExceptionHandler = null;

  public JobHandlerThreadFactoryBuilder setNamePrefix(String namePrefix) {
    if (namePrefix == null) {
      throw new NullPointerException();
    }
    this.namePrefix = namePrefix;
    return this;
  }

  public JobHandlerThreadFactoryBuilder setDaemon(boolean daemon) {
    this.daemon = daemon;
    return this;
  }

  public JobHandlerThreadFactoryBuilder setPriority(int priority) {
    if (priority < Thread.MIN_PRIORITY) {
      throw new IllegalArgumentException(
          String.format("Thread priority (%s) must be >= %s", priority, Thread.MIN_PRIORITY));
    }

    if (priority > Thread.MAX_PRIORITY) {
      throw new IllegalArgumentException(
          String.format("Thread priority (%s) must be <= %s", priority, Thread.MAX_PRIORITY));
    }

    this.priority = priority;
    return this;
  }

  public JobHandlerThreadFactoryBuilder setUncaughtExceptionHandler(UncaughtExceptionHandler uncaughtExceptionHandler) {
    if (null == uncaughtExceptionHandler) {
      throw new NullPointerException("UncaughtExceptionHandler cannot be null");
    }
    this.uncaughtExceptionHandler = uncaughtExceptionHandler;
    return this;
  }

  public JobHandlerThreadFactoryBuilder setThreadFactory(ThreadFactory backingThreadFactory) {
    if (null == uncaughtExceptionHandler) {
      throw new NullPointerException("BackingThreadFactory cannot be null");
    }
    this.backingThreadFactory = backingThreadFactory;
    return this;
  }

  public ThreadFactory build() {
    return build(this);
  }

  private static ThreadFactory build(JobHandlerThreadFactoryBuilder builder) {
    final String namePrefix = builder.namePrefix;
    final Boolean daemon = builder.daemon;
    final Integer priority = builder.priority;
    final UncaughtExceptionHandler uncaughtExceptionHandler = builder.uncaughtExceptionHandler;
    final ThreadFactory backingThreadFactory = (builder.backingThreadFactory != null) ? builder.backingThreadFactory
        : Executors.defaultThreadFactory();

    final AtomicLong count = new AtomicLong(0);

    return new ThreadFactory() {
      @Override
      public Thread newThread(Runnable runnable) {
        Thread thread = backingThreadFactory.newThread(runnable);
        if (namePrefix != null) {
          thread.setName(namePrefix + "-" + count.getAndIncrement());
        }
        if (daemon != null) {
          thread.setDaemon(daemon);
        }
        if (priority != null) {
          thread.setPriority(priority);
        }
        if (uncaughtExceptionHandler != null) {
          thread.setUncaughtExceptionHandler(uncaughtExceptionHandler);
        }
        return thread;
      }
    };
  }
}